'' System Version 2.002
'' OS_keyboard Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/11/01 - Changed to Hydra driver with Propeller OS mods
''   2006/11/04 - Moved stop routine from OS_keyboardInit to here

'' This is the PS/2 keyboard driver for the Propeller OS.  It is based on the PS/2 Keyboard Driver
'' (Isolated) v1.0 (C) 2005 Parallax, Inc. which was modified primarily to support a dynamically
'' allocated work area.  This portion is intended to be incorporated into programs and is the
'' keyboard application interface.  Note that this part of the keyboard driver can work with either
'' OS_keyboardInit version (for the Propeller Demo Board or the Hydra Game System).

CON
  kbd_tail    =  0       ' Key buffer tail       read/write (20 contiguous longs)
  kbd_head    =  4       ' Key buffer head       read-only
  kbd_present =  8       ' Keyboard present      read-only
  kbd_states  = 12       ' Key states (256 bits) read-only
  kbd_keys    = 44       ' Key buffer (16 words) read-only  (used for initial parameters)
  kbd_cog     = 76       ' Key cog bitmask       read-only

OBJ
  ldr : "OS_Loader"

PUB stop | p, c

'' Stop keyboard driver - frees a cog

  p := long[ldr#keybdPtr]
  c := long[p+kbd_cog]
  if c
    cogstop(c >> 8)
  long[p+kbd_cog] := 0
  ldr.unregisterCOG(c & $FF)
  longfill(p+kbd_tail, 0, 20)

PUB present

'' Check if keyboard present - valid ~2s after start - returns true/false

  return -long[long[ldr#keybdPtr]+kbd_present]

PUB key : keycode | p, t

'' Get key (never waits)
'' returns key (0 if buffer empty)

  p := long[ldr#keybdPtr]
  t := long[p+kbd_tail]
  if t <> long[p+kbd_head]
    keycode := word[p+kbd_keys][t]
    long[p+kbd_tail] := ++t & $F

PUB getkey : keycode

'' Get next key (may wait for keypress)
'' returns key

  repeat until (keycode := key)
    ?long[long[ldr#systemPtr]+ldr#sys_random]

PUB newkey

'' Clear buffer and get new key (always waits for keypress)
'' returns key

  clearkeys
  return getkey

PUB gotkey | p

'' Check if any key in buffer - returns true/false

  p := long[ldr#keybdPtr]
  return long[p+kbd_tail] <> long[p+kbd_head]

PUB clearkeys | p

'' Clear key buffer

  p := long[ldr#keybdPtr]
  long[p+kbd_tail] := long[p+kbd_head]

PUB keystate(k)

'' Get the state of a particular key - returns true/false

  return -(long[long[ldr#keybdPtr]+kbd_states][k >> 5] >> k & 1)

''      _________
''      Key Codes
''
''      00..DF  = keypress and keystate
''      E0..FF  = keystate only
''
''
''      09      Tab
''      0D      Enter
''      20      Space
''      21      !
''      22      "
''      23      #
''      24      $
''      25      %
''      26      &
''      27      '
''      28      (
''      29      )
''      2A      *
''      2B      +
''      2C      ,
''      2D      -
''      2E      .
''      2F      /
''      30      0..9
''      3A      :
''      3B      ;
''      3C      <
''      3D      =
''      3E      >
''      3F      ?
''      40      @       
''      41..5A  A..Z
''      5B      [
''      5C      \
''      5D      ]
''      5E      ^
''      5F      _
''      60      `
''      61..7A  a..z
''      7B      {
''      7C      |
''      7D      }
''      7E      ~
''
''      80-BF   (future international character support)
''
''      C0      Left Arrow
''      C1      Right Arrow
''      C2      Up Arrow
''      C3      Down Arrow
''      C4      Home
''      C5      End
''      C6      Page Up
''      C7      Page Down
''      C8      Backspace
''      C9      Delete
''      CA      Insert
''      CB      Esc
''      CC      Apps
''      CD      Power
''      CE      Sleep
''      CF      Wakeup
''
''      D0..DB  F1..F12
''      DC      Print Screen
''      DD      Scroll Lock
''      DE      Caps Lock
''      DF      Num Lock
''
''      E0..E9  Keypad 0..9
''      EA      Keypad .
''      EB      Keypad Enter
''      EC      Keypad +
''      ED      Keypad -
''      EE      Keypad *
''      EF      Keypad /
''
''      F0      Left Shift
''      F1      Right Shift
''      F2      Left Ctrl
''      F3      Right Ctrl
''      F4      Left Alt
''      F5      Right Alt
''      F6      Left Win
''      F7      Right Win
''
''      FD      Scroll Lock State
''      FE      Caps Lock State
''      FF      Num Lock State
''
''      +100    if Shift
''      +200    if Ctrl
''      +400    if Alt
''      +800    if Win
''
''      eg. Ctrl-Alt-Delete = $6C9
''
''
'' Note: Driver will buffer up to 15 keystrokes, then ignore overflow.